from setuptools import setup


setup(
    name='irl2019',
    version='0.4',
    description='Temporary package to support an event.',
    url='',
    author='Matthew Honnibal',
    author_email='matt@explosion.ai',
    license='MIT',
    packages=['irl2019'],
    zip_safe=False
)
